/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2009-2011 ev-i GmbH & IBCL e.U.
 * Copyright (C) 2011-2013 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCYCHANNEL_H_
#define FANCYCHANNEL_H_

#include <fancyobject.h>

namespace Fancy {

class Map;

enum ShutdownOption {

    ShutdownRead  = 0,
    ShutdownWrite = 1,
    ShutdownBoth  = 2
};

/*!
 * A read and writable I/O channel like a stream or a file.
 */
class FANCYBASE_API Channel: public virtual Fancy::Object {
public:
    typedef Fancy::ObjPtr<Channel> ptr_t;

    /**
     * Wrap the given channel with a channel, that issues buffered reads and writes
     * on the underlying channel.
     */
    static ptr_t newBufferedInstance(const Channel::ptr_t channel, int bufsize);

    /**
     * Unconditionally read @c len bytes from the channel into @c data.
     * If @c max_len is greater than len, allow reading up to @max_len bytes.
     * It may be needed to call barebone read() multiple times. If the
     * stream ends prematurely, an exception is thrown unless
     * @c max_len is set to zero.
     * @returns The number of bytes actually being read, which is always len, if
     *          @c max_len is set to -1. The returned number may be
     *          smaller then @c len if and only if @c max_len is set to 0
     *          and the stream ends prematurely. If @c max_len is greater than
     *          @len, the returned number is greater or equal than @len.
     *
     */
    static int read_all(const Channel::ptr_t &channel, uint8_t *data, int len, int max_len = -1);

    /**
     * Unconditionally write @c len bytes from the channel into @c data.
     * It may be needed to call barebone write() multiple times. If the
     * stream ends prematurely, an exception is thrown.
     */
    static void write_all(const Channel::ptr_t &channel, const uint8_t *data, int len);

    /*!
     * Read a chunk of data from the channel.
     */
    virtual int read (void *buf, int count) = 0;

    /*!
     * Write a chunk of data to the channel.
     */
    virtual int write (const void *buf, int count) = 0;

    virtual void flush() = 0;

    virtual void close() = 0;

    virtual bool isOpen() const = 0;

    /**
     * Shutdown the communication on the channel before actually closing it.
     * In multiprocess enivronments, it may be feasible to not
     * call shutdown from a file descriptor inherited by a child process,
     * because this way the parent process may be interrupted, too.
     * On the other hand, in multi-threaded environments, calling shutdown
     * is a good opportunity to cancel any blocking operation in a concurrent
     * thread.
     *
     * @param o Whether to shut down read, write or all communication.
     */
    virtual void shutdown(ShutdownOption o) = 0;

protected:
    Channel() {}
    virtual ~Channel();
};

}

#endif /* FANCYCHANNEL_H_ */
