/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2009-2011 ev-i GmbH & IBCL e.U.
 * Copyright (C) 2011-2013 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCYNUMBERS_H_
#define FANCYNUMBERS_H_

#include <fancyserializable.h>

namespace Fancy {

class FANCYBASE_API Number: public Serializable {
public:
    typedef ObjPtr<Number> ptr_t;

    Number() {}

    virtual int8_t  byteValue()   const = 0;
    virtual int16_t shortValue()  const = 0;
    virtual int32_t intValue()    const = 0;
    virtual int64_t longValue()   const = 0;
    virtual float   floatValue()  const = 0;
    virtual double  doubleValue() const = 0;

    // these functions are here in order to fulfill the requirements of
    // ArrayObjectList<Number> upon serialization and deserialization.
    static int64_t getSerialVersionUID_s();
    static const char *getSerialID_s();

protected:
    virtual ~Number();
};

class FANCYBASE_API Byte: public Number {

    int8_t v;

public:
    typedef ObjPtr<Byte> ptr_t;
    typedef int8_t native_t;

    static ObjPtr<Byte>   newInstance(const std::string &s);
    static ObjPtr<Byte>   newInstance(int8_t _v);
    static ObjPtr<Serializable> newSerializableInstance();
    static ObjPtr<Byte>   newInstance(const Object::ptr_t &obj);

    static int8_t toByte(const Object::ptr_t &obj);
    static int8_t nativeFromObject(const Object::ptr_t &obj) {
    	return Byte::toByte(obj);
    }
    static int8_t toByte(const std::string &s);
    static std::string toString(int8_t i);

    static int getBitLength(int8_t i);

    int8_t  byteValue()   const;
    int16_t shortValue()  const;
    int32_t intValue()    const;
    int64_t longValue()   const;
    float   floatValue()  const;
    double  doubleValue() const;

    std::string toString() const;

    bool lessThan(const Object &o) const;
    bool equals(const Object &o) const;

    // implementation of Fancy::Serializable
    static int64_t getSerialVersionUID_s();
    static const char *getSerialID_s();
    static const char *getSuperSerialID_s();

    int64_t getSerialVersionUID() const;
    const char *getSerialID() const;
    const char *getSuperSerialID() const;
    void writeFields(const ObjPtr<Serializer> &serializer) const;
    ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);

protected:
    Byte(int8_t _v) :v(_v) {}
    ~Byte();
};

class FANCYBASE_API Short: public Number {

    int16_t v;

public:
    typedef ObjPtr<Short> ptr_t;
    typedef int16_t native_t;

    static ObjPtr<Short>   newInstance(const std::string &s);
    static ObjPtr<Short>   newInstance(int16_t _v);
    static ObjPtr<Serializable> newSerializableInstance();
    static ObjPtr<Short>   newInstance(const Object::ptr_t &obj);

    static int16_t toShort(const Object::ptr_t &obj);
    static int16_t nativeFromObject(const Object::ptr_t &obj) {
    	return Short::toShort(obj);
    }
    static int16_t toShort(const std::string &s);
    static std::string toString(int16_t i);

    static int getBitLength(int16_t i);

    int8_t  byteValue()   const;
    int16_t shortValue()  const;
    int32_t intValue()    const;
    int64_t longValue()   const;
    float   floatValue()  const;
    double  doubleValue() const;

    std::string toString() const;

    bool lessThan(const Object &o) const;
    bool equals(const Object &o) const;

    // implementation of Fancy::Serializable
    static int64_t getSerialVersionUID_s();
    static const char *getSerialID_s();
    static const char *getSuperSerialID_s();

    int64_t getSerialVersionUID() const;
    const char *getSerialID() const;
    const char *getSuperSerialID() const;
    void writeFields(const ObjPtr<Serializer> &serializer) const;
    ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);

protected:
    Short(int16_t _v) :v(_v) {}
    ~Short();
};

class FANCYBASE_API Integer: public Number {

    int32_t v;

public:
    typedef ObjPtr<Integer> ptr_t;
    typedef int32_t native_t;

    static ObjPtr<Integer> newInstance(const std::string &s);
    static ObjPtr<Integer> newInstance(int32_t _v);
    static ObjPtr<Serializable> newSerializableInstance();
    static ObjPtr<Integer> newInstance(const Object::ptr_t &obj);

    static int32_t toInteger(const Object::ptr_t &obj);
    static int32_t nativeFromObject(const Object::ptr_t &obj) {
    	return Integer::toInteger(obj);
    }
    static int32_t toInteger(const std::string &s);
    static std::string toString(int32_t i);

    static int getBitLength(int32_t i);

    int8_t  byteValue()   const;
    int16_t shortValue()  const;
    int32_t intValue()    const;
    int64_t longValue()   const;
    float   floatValue()  const;
    double  doubleValue() const;

    std::string toString() const;

    bool lessThan(const Object &o) const;
    bool equals(const Object &o) const;

    // implementation of Fancy::Serializable
    static int64_t getSerialVersionUID_s();
    static const char *getSerialID_s();
    static const char *getSuperSerialID_s();

    int64_t getSerialVersionUID() const;
    const char *getSerialID() const;
    const char *getSuperSerialID() const;
    void writeFields(const ObjPtr<Serializer> &serializer) const;
    ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);

protected:
    Integer(int32_t _v) :v(_v) {}
    ~Integer();
};

class FANCYBASE_API Long: public Number {

    int64_t v;

public:
    typedef ObjPtr<Long> ptr_t;
    typedef int64_t native_t;

    static ObjPtr<Long>   newInstance(const std::string &s);
    static ObjPtr<Long>   newInstance(int64_t _v);
    static ObjPtr<Serializable> newSerializableInstance();
    static ObjPtr<Long>   newInstance(const Object::ptr_t &obj);

    static int64_t toLong(const Object::ptr_t &obj);
    static int64_t nativeFromObject(const Object::ptr_t &obj) {
    	return Long::toLong(obj);
    }
    static int64_t toLong(const std::string &s);
    static std::string toString(int64_t i);

    static int getBitLength(int64_t i);

    int8_t  byteValue()   const;
    int16_t shortValue()  const;
    int32_t intValue()    const;
    int64_t longValue()   const;
    float   floatValue()  const;
    double  doubleValue() const;

    std::string toString() const;

    bool lessThan(const Object &o) const;
    bool equals(const Object &o) const;

    // implementation of Fancy::Serializable
    static int64_t getSerialVersionUID_s();
    static const char *getSerialID_s();
    static const char *getSuperSerialID_s();

    int64_t getSerialVersionUID() const;
    const char *getSerialID() const;
    const char *getSuperSerialID() const;
    void writeFields(const ObjPtr<Serializer> &serializer) const;
    ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);

protected:
    Long(int64_t _v) :v(_v) {}
    ~Long();
};

class FANCYBASE_API Float: public Number {

    float v;

public:
    typedef ObjPtr<Float> ptr_t;
    typedef float native_t;

    static ObjPtr<Float>   newInstance(const std::string &s);
    static ObjPtr<Float>   newInstance(float _v);
    static ObjPtr<Serializable> newSerializableInstance();
    static ObjPtr<Float>   newInstance(const Object::ptr_t &obj);

    static float toFloat(const Object::ptr_t &obj);
    static float nativeFromObject(const Object::ptr_t &obj) {
    	return Float::toFloat(obj);
    }
    static float toFloat(const std::string &s);
    static std::string toString(float f);

    int8_t  byteValue()   const;
    int16_t shortValue()  const;
    int32_t intValue()    const;
    int64_t longValue()   const;
    float   floatValue()  const;
    double  doubleValue() const;

    std::string toString() const;

    bool lessThan(const Object &o) const;
    bool equals(const Object &o) const;

    // implementation of Fancy::Serializable
    static int64_t getSerialVersionUID_s();
    static const char *getSerialID_s();
    static const char *getSuperSerialID_s();

    int64_t getSerialVersionUID() const;
    const char *getSerialID() const;
    const char *getSuperSerialID() const;
    void writeFields(const ObjPtr<Serializer> &serializer) const;
    ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);

protected:
    Float(float _v) :v(_v) {}
    ~Float();
};

class FANCYBASE_API Double: public Number {

    double v;

public:
    typedef ObjPtr<Double> ptr_t;
    typedef double native_t;

    static ObjPtr<Double>   newInstance(const std::string &s);
    static ObjPtr<Double>   newInstance(double _v);
    static ObjPtr<Serializable> newSerializableInstance();
    static ObjPtr<Double>   newInstance(const Object::ptr_t &obj);

    static double toDouble(const Object::ptr_t &obj);
    static double nativeFromObject(const Object::ptr_t &obj) {
    	return Double::toDouble(obj);
    }
    static double toDouble(const std::string &s);
    static std::string toString(double d);

    int8_t  byteValue()   const;
    int16_t shortValue()  const;
    int32_t intValue()    const;
    int64_t longValue()   const;
    float   floatValue()  const;
    double  doubleValue() const;

    std::string toString() const;

    bool lessThan(const Object &o) const;
    bool equals(const Object &o) const;

    // implementation of Fancy::Serializable
    static int64_t getSerialVersionUID_s();
    static const char *getSerialID_s();
    static const char *getSuperSerialID_s();

    int64_t getSerialVersionUID() const;
    const char *getSerialID() const;
    const char *getSuperSerialID() const;
    void writeFields(const ObjPtr<Serializer> &serializer) const;
    ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);

protected:
    Double(double _v) :v(_v) {}
    ~Double();
};

class FANCYBASE_API Boolean: public Serializable {

    bool v;

public:
    typedef ObjPtr<Boolean> ptr_t;
    typedef bool native_t;

    static ObjPtr<Boolean> newInstance(const std::string &s);
    static ObjPtr<Boolean> newInstance(bool _v);
    static ObjPtr<Serializable>  newSerializableInstance();

    static bool toBoolean(const Object::ptr_t &obj);
    static bool nativeFromObject(const Object::ptr_t &obj) {
    	return Boolean::toBoolean(obj);
    }
    static bool toBoolean(const std::string &s);
    static std::string toString(bool b);

    bool booleanValue()   const;

    std::string toString() const;

    bool lessThan(const Object &o) const;
    bool equals(const Object &o) const;

    // implementation of Fancy::Serializable
    static int64_t getSerialVersionUID_s();
    static const char *getSerialID_s();

    int64_t getSerialVersionUID() const;
    const char *getSerialID() const;
    void writeFields(const ObjPtr<Serializer> &serializer) const;
    ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);

protected:
    Boolean(bool _v) :v(_v) {}
    ~Boolean();
};

class FANCYBASE_API Character: public Serializable {

    wchar_t v;

public:
    typedef ObjPtr<Character> ptr_t;
    typedef wchar_t native_t;

    static ObjPtr<Character> newInstance(const std::string &s);
    static ObjPtr<Character> newInstance(wchar_t _v);
    static ObjPtr<Serializable>  newSerializableInstance();

    static wchar_t toCharacter(const Object::ptr_t &obj);
    static wchar_t nativeFromObject(const Object::ptr_t &obj) {
     	return Character::toCharacter(obj);
    }
    static wchar_t toCharacter(const std::string &s);
    static std::string toString(wchar_t c);

    wchar_t charValue() const;

    std::string toString() const;

    bool lessThan(const Object &o) const;
    bool equals(const Object &o) const;

    // implementation of Fancy::Serializable
    static int64_t getSerialVersionUID_s();
    static const char *getSerialID_s();

    int64_t getSerialVersionUID() const;
    const char *getSerialID() const;
    void writeFields(const ObjPtr<Serializer> &serializer) const;
    ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);

protected:
    Character(wchar_t _v) :v(_v) {}
    ~Character();
};
}

#endif /* FANCYNUMBERS_H_ */
