/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2009-2011 ev-i GmbH & IBCL e.U.
 * Copyright (C) 2011-2013 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

/****************************************************************************
**
** Author: Wolfgang Glas / EV-i
**
** A thread-safe reference counted object.
**
*****************************************************************************/

#ifndef FANCYBASE_OBJECT_H
#define FANCYBASE_OBJECT_H

#include <fancyint.h>
#include <fancybase.h>
#include <typeinfo>
#include <iosfwd>
#ifdef __GNUG__
#include <bits/atomic_word.h>
#endif

namespace Fancy {

#ifdef WIN32
  typedef long ref_count_t;
#elif defined ( __GNUG__ )
  typedef _Atomic_word ref_count_t;
#else
  typedef unsigned ref_count_t;
#endif

  template <class T> class ObjPtr;

  class FANCYBASE_API Object
  {
    ref_count_t nref;

  public:
    typedef Fancy::ObjPtr<Object> ptr_t;

    Object() : nref(0) {}

    /** Increment reference count by one. */
    void ref() throw();

    /** Decrement reference count by one.
        If reference count drops to zero, delete this and return true. */
    bool deref() throw();

    /**
     * Return a string representation of this object. The default implementation
     * return the pointer and the name of the class of this instance.
     */
    virtual std::string toString() const;

    /*!
     * @return A unique ID for this version of the object.
     */
    virtual int64_t getSerialVersionUID() const;

    /*!
     * @return The name of this object, which usually corresponds to a JAVA
     *         class name. If this class is not serializable, the C++ RTTI
     *         typeid is returned.
     */
    virtual const char *getSerialID() const;

    /*!
     * @return The name of the base class of this object,
     *         which usually corresponds to a JAVA
     *         class name. If this class is not serializable, or the
     *         super class is a barebone Object, NULL is returned.
     */
    virtual const char *getSuperSerialID() const;

    /**
     * A comparison operation, which should be implemented by each implementation
     * of @c Fancy::Object.
     *
     * At typical implementation, which exposes the specializes less than operator
     * looks like:
     * @verbatim
     * bool UID::lessThan(const Object &o) const {
     *   const UID *uid = dynamic_cast<const UID*>(&o);
     *   if (!uid) return this->uncomparableLessThan(o);
     *   return *this < *uid;
     * }
     * @endverbatim
     *
     * This function should be implemented in a strictly reflexive way, assuring
     * that @c a.lessThan(b) implies @c b.lessThan(a). Moreover, both
     * @c a.lessThan(b) and @c b.lessThan(a) should simultaneously return
     * @c false if and only if @c a and @c b are equal.
     *
     * @see Fancy::Object::uncomparableLessThan(const Fancy::Object &)
     */
    virtual bool lessThan(const Object &o) const;

    /**
     * A comparison operation, which should be implemented by each implementation
     * of @c Fancy::Object.
     *
     * It is permissible to not implement @c equals, because the default implementation
     * returns
     * @verbatim
     *   !this->lessThan(o) && !o.lessThan(*this)
     * @endverbatim
     *
     * However, it is encouraged to implement @c equals alongside with @c lessThan, mainly
     * for performace reasons.
     */
    virtual bool equals(const Object &o) const;

    // these functions are here in order to fulfill the requirements of
    // ArrayObjectList<Object> upon serialization and deserialization.
    static int64_t getSerialVersionUID_s();
    static const char *getSerialID_s();
    static const char *getSuperSerialID_s();

  protected:
    virtual ~Object();

    /**
     * A helper function for comparing two object, which are not comparable
     * otherwise. This method simply compares the class names of @c this and
     * @c o.
     *
     * @see Fancy::Object::lessThan(const Fancy::Object &)
     */
    bool uncomparableLessThan(const Object &o) const;

  private:
    Object(const Object &a);
    Object &operator= (const Object &a);
  };

  FANCYBASE_API std::ostream& operator<<(std::ostream& os, const Object &x);

  FANCYBASE_API std::ostream& writeObjectPtr(std::ostream& os, const Object* x);

  /**
   * If x is NULL, returns "null".
   * Otherwise returns x->toString()
   * Fancy::String, Fancy::Qt4::String and most containers reimplement Fancy::Object::toString() just as one hopes.
   */
  FANCYBASE_API std::string objectPtrToString(const Object* x);

  /**
   * Throws a Fancy::Exception
   */
  FANCYBASE_API void throwNull(const char *fmt, ...)
#ifdef __GNUG__
  __attribute__((format(printf,1,2)))
#endif
  ;

  /**
   * If p.isNull(), delivers "null".
   * Otherwise delivers p->toString().
   * Fancy::String, Fancy::Qt4::String and most containers reimplement Fancy::Object::toString() just as one hopes.
   */
#define FANCY_OBJ_PTR_STRING(p) (::Fancy::objectPtrToString((p).get()))

  /**
   * Pass an ObjPtr instance to a %s format specifier.
   * If p.isNull(), delivers "null".
   * Otherwise delivers p->toString().c_str()
   * Fancy::String, Fancy::Qt4::String and most containers reimplement Fancy::Object::toString() just as one hopes.
   */
#define FANCY_OBJ_PTR_ARG(p) (::Fancy::objectPtrToString((p).get()).c_str())

  /**
   * Throws an exception with source code file and line if p.isNull()
   */
#define FANCY_OBJ_PTR_CHECK(p) do { if ((p).isNull()) ::Fancy::throwNull("%s:%d(%s): Pointer to type %s is null.",\
        __FILE__,__LINE__,__FUNCTION__,(p).getTargetTypeName()); } while(0)

  // A pointer to a subclass of Object. For a non-intrusive approach see
  // Fancy::Ptr<T> in fancyptr.h
  template <class T> class ObjPtr
  {
    T *ref;

  public:
    typedef T target_t;

    ObjPtr() : ref(0) {}
    ObjPtr(T *_t) : ref(_t) { if (ref) ref->ref(); }

    template<class U>ObjPtr(const ObjPtr<U> &a) : ref(dynamic_cast<T*>(a.get()))
    { if (ref) ref->ref(); }

    ObjPtr(const ObjPtr &a) : ref(a.ref) { if (ref) ref->ref(); }

    ObjPtr &operator=(const ObjPtr &a) {
      if (ref) ref->deref();
      ref = a.ref;
      if (ref) ref->ref();
      return *this;
    }

    template<class U>ObjPtr &operator=(const ObjPtr<U> &a) {
      if (ref) ref->deref();
      ref=dynamic_cast<T*>(a.get());
      if (ref) ref->ref();
      return *this;
    }

    ObjPtr &operator=(T *_t) {
      if (ref) ref->deref();
      ref=_t;
      if (ref) ref->ref();
      return *this;
    }

    template<class U>ObjPtr<U> check_cast() const {
        ObjPtr<U> r(*this);
        if (!this->isNull())
            FANCY_OBJ_PTR_CHECK(r);
        return r;
    }

    void setNull() { if (ref) ref->deref(); ref=0; }

    ~ObjPtr() { if (ref) ref->deref(); }

    bool isNull() const { return ref == 0; }

    T *operator->() const { return ref; }
    T &operator*() const { return *ref; }

    /*!
       Indirection of pointer.
       This method should normally not be used, because
       the lifetime of returned pointer is tied to the lifetime of this
       Ptr object.
     */
    T *get() const { return ref; }

    const char *getTargetTypeName() const { return typeid(target_t).name(); }
    const char *getTargetSerialID() const { return target_t::getSerialID_s(); }
    int64_t getTargetSerialVersionUID() const { return target_t::getSerialVersionUID_s(); }

    bool operator!() const {
      return ref == 0;
    }

    template<class U> bool operator== (const ObjPtr<U> &a) const {
      return a.get() == this->get();
    }

    template<class U> bool operator!= (const ObjPtr<U> &a) const {
      return a.get() != this->get();
    }

    template<class U> bool operator< (const ObjPtr<U> &a) const {
      return a.get() < this->get();
    }
  };

  /**
   * This function pointer type is the prototype of the static
   * @c newObjInstance method which should be present in all Object implementations.
   */
  typedef ObjPtr<Object> (*ObjPtr_newInstance_t)();

  /**
   * The functions type, which is used as comparison operator in
   * STL containers, which hold @c ObjPtr<Object> instances.
   *
   * You hardly ever need to implement a comparator for @c ObjPtr<Object>
   * other than @c FAncy::fwd_less.
   *
   * @param a The first element to compare or @c NULL.
   * @param b The second element to compare or @c NULL.
   * @return Whether a is less than b.
   * @see Fancy::fwd_less
   */
  typedef bool (*ObjPtr_less_t)(const ObjPtr<Object> &a, const ObjPtr<Object> &b);

  /**
   * An instance of @c ObjPtr_less_t, which is implemented like follows:
   *
   * @verbatim
   *  bool fwd_less(const ObjPtr<Object> &a, const ObjPtr<Object> &b) {
   *     if (a.isNull()) return !b.isNull();
   *     if (b.isNull()) return false;
   *     return a->lessThan(*b);
   *  }
   * @endverbatim
   *
   * The semantics of this operator is, the @c NULL is the smaller than an non-null
   * object pointer. For non-null pointers, the comparison is dispatched to
   * @c Fancy::Objecct::lessThan(const Object &o).
   */
  FANCYBASE_API bool fwd_less(const ObjPtr<Object> &a, const ObjPtr<Object> &b);

  /**
   * An instance of @c ObjPtr_less_t, which is implemented like follows:
   *
   * @verbatim
   *  bool fwd_equals(const ObjPtr<Object> &a, const ObjPtr<Object> &b) {
   *     if (a.isNull()) return b.isNull();
   *     if (b.isNull()) return false;
   *     return a->equals(*b);
   *  }
   * @endverbatim
   */
  FANCYBASE_API bool fwd_equals(const ObjPtr<Object> &a, const ObjPtr<Object> &b);

  template<class T>
  inline std::ostream& operator<<(std::ostream& os, const ObjPtr<T> &p) {

    return writeObjectPtr(os,p.get());
  }

}

#endif
