/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2009-2011 ev-i GmbH & IBCL e.U.
 * Copyright (C) 2011-2013 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCY_POLLHANDLERS_H_
#define FANCY_POLLHANDLERS_H_

#include <fancypollablechannel.h>

#include <fancyarray.h>
#include <fancyfilehandle.h>

#include <string>


namespace Fancy {


/**
 * A basic PollEventHandler implementation.
 * Polled data are just swallowed.
 */
class FANCYBASE_API BasicPollHandler : public virtual PollEventHandler {

public:
    typedef Fancy::ObjPtr<BasicPollHandler> ptr_t;

    virtual void handleEvent(
            const Fancy::ObjPtr<Fancy::ChannelPoller> &context,
            const Fancy::ObjPtr<Fancy::PollableChannel> &channel,
            Fancy::PollEvent event);

    static ptr_t newInstance(const char* name);

protected:
    BasicPollHandler(const char* name);
    virtual ~BasicPollHandler();

private:
    const char* name;

};


/**
 * A data-collecting PollEventHandler implementation, i.e. for collecting outputs of called commands' stdout and/or stderr.
 * Polled data are collected and can be retrieved using getCollectedBytes().
 */
class FANCYBASE_API CollectingReadPollHandler : public virtual PollEventHandler {

public:
    typedef Fancy::ObjPtr<CollectingReadPollHandler> ptr_t;

    static ptr_t newInstance(
            const char* name,
            size_t collectionBufferSize);

    virtual void handleEvent(
            const Fancy::ObjPtr<Fancy::ChannelPoller> &context,
            const Fancy::ObjPtr<Fancy::PollableChannel> &channel,
            Fancy::PollEvent event);

    virtual size_t getNrCollectedBytes();
    virtual const Fancy::Array<char> & getCollectedBytes();

protected:
    CollectingReadPollHandler(
            const char* name,
            size_t _collectionBufferSize);
    virtual ~CollectingReadPollHandler();

private:
    const char* name;

    size_t bufferSize;
    Fancy::Array<char> buffer;

    size_t collectionSize;
    Fancy::Array<char> collectedData;
    size_t nrBytesCollected;

};


/**
 * Callback interface for LineReadingPollHandler (see below).
 */
class FANCYBASE_API LineReadingPollHandlerCallback : public virtual Object {

public:
    typedef Fancy::ObjPtr<LineReadingPollHandlerCallback> ptr_t;

    // Called everytime a complete line was received. Line does not contain any trailing \r or \n.
    virtual void onLinePolled(const std::string &line) = 0;
};


/**
 * A line-reading PollEventHandler implementation, i.e. for line-wise retrieval of outputs of called commands' stdout and/or stderr.
 * Incomplete lines are not forwarded, which may produce nasty behaviour with commands not completing their last output line.
 */
class FANCYBASE_API LineReadingPollHandler : public virtual PollEventHandler {

public:
    typedef Fancy::ObjPtr<LineReadingPollHandler> ptr_t;

    static ptr_t newInstance(
            const char* name,
            const LineReadingPollHandlerCallback::ptr_t &linesReceiver,
            size_t collectionBufferSize=16*1024);

    virtual void handleEvent(
            const Fancy::ObjPtr<Fancy::ChannelPoller> &context,
            const Fancy::ObjPtr<Fancy::PollableChannel> &channel,
            Fancy::PollEvent event);

protected:
    LineReadingPollHandler(
            const char* name,
            const LineReadingPollHandlerCallback::ptr_t &linesReceiver,
            size_t collectionBufferSize=16*1024);
    virtual ~LineReadingPollHandler();

private:
    const char* name;
    LineReadingPollHandlerCallback::ptr_t linesReceiver;

    size_t bufferSize;
    Fancy::Array<char> buffer;
    size_t writePos;

};



class FANCYBASE_API ForwardingReadPollHandler : public Fancy::PollEventHandler {

public:
    typedef Fancy::ObjPtr<ForwardingReadPollHandler> ptr_t;

    static ptr_t newInstance(
            const char* name,
            const Fancy::FileHandle::ptr_t out);

    virtual void handleEvent(
            const Fancy::ObjPtr<Fancy::ChannelPoller> &context,
            const Fancy::ObjPtr<Fancy::PollableChannel> &channel,
            Fancy::PollEvent event);

protected:
    ForwardingReadPollHandler(
            const char* name,
            const Fancy::ObjPtr<Fancy::FileHandle> out);

    virtual ~ForwardingReadPollHandler();

private:
    const char* name;
    Fancy::FileHandle::ptr_t out;
    size_t dataSize;
    Fancy::Array<uint8_t> data;

};

/**
 * A poll handler to send a string to a descriptor,
 * might be used to send a possibly multi-line string to
 * stdin of a forked process.
 *
 * A SendStringPollhandle should be registered with @c PollOut interest.
 */
class FANCYBASE_API SendStringPollHandler : public virtual Fancy::PollEventHandler {

public:
    typedef Fancy::ObjPtr<SendStringPollHandler> ptr_t;

    /**
     * Create a new instance, which will send
     * @c data to the channel.
     *
     * If @c closeOnDataSent is true, close the channel,
     * otherwise just change the channel interest from @c PollOut
     * to @c PollNone
     */
    static ptr_t newInstance(const std::string &data,
            bool closeOnDataSent);

    virtual void handleEvent(
            const Fancy::ObjPtr<Fancy::ChannelPoller> &context,
            const Fancy::ObjPtr<Fancy::PollableChannel> &channel,
            Fancy::PollEvent event);

protected:
    SendStringPollHandler(const std::string &_data, bool _closeOnDataSent);

    virtual ~SendStringPollHandler();

private:
    std::string data;
    size_t bytesSent;
    bool closeOnDataSent;
};


}

#endif /* FANCY_POLLHANDLERS_H_ */
