/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2009-2011 ev-i GmbH & IBCL e.U.
 * Copyright (C) 2011-2013 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

/****************************************************************************
**
** Author: Wolfgang Glas / EV-i
**
** Wrapper calsses for pthread functions.
**
*****************************************************************************/

#ifndef FANCYBASE_THREAD_H
#define FANCYBASE_THREAD_H

#include <fancyrunnable.h>

#ifdef WIN32
typedef void *pthread_mutex_t;
namespace Fancy { class WaitConditionPrivate; }
typedef Fancy::WaitConditionPrivate *pthread_cond_t;
typedef void *pthread_t;
typedef long pthread_once_t;
#define PTHREAD_ONCE_INIT 0
#else
#include <pthread.h>
#endif

namespace Fancy {

  class ThreadCallback;

  class FANCYBASE_API Mutex
  {
  public:
    Mutex() throw();
    ~Mutex() throw();

    void lock() throw();
    void unlock() throw();

    pthread_mutex_t *handle() throw();
  private:
    Mutex(const Mutex &);
    Mutex &operator= (const Mutex &);

    pthread_mutex_t mutex;
  };

  /**
   * One-time initialization.
   */
  class FANCYBASE_API Once
  {
  public:
    Once() : once(PTHREAD_ONCE_INIT) {}
    ~Once() {}

    /**
     * Call the given function once for this instance of @c Once.
     *
     * @param init_rountine The subroutine to call once.
     */
    void runOnce(void (*init_routine)(void));
  private:
    Once(const Once &);
    Once &operator= (const Once &);

    pthread_once_t once;
  };


  class FANCYBASE_API FileMutex
  {
  public:
    FileMutex();
    ~FileMutex();

    void open(const char *fn);

    void lock();
    void unlock();

  private:
    FileMutex(const FileMutex &);
    FileMutex &operator= (const FileMutex &);

    int fd;
  };

  template<class MutexT>
  class BasicMutexLocker
  {
    MutexT *mutex;
  public:
    BasicMutexLocker(MutexT *m) : mutex(m) { if (mutex) mutex->lock(); }
    ~BasicMutexLocker() { if (mutex) mutex->unlock(); }
    void unlock() { if (mutex) { mutex->unlock(); mutex=0; } }
    void lock(MutexT *m)
    { if (mutex) mutex->unlock(); mutex=m; if (mutex) mutex->lock(); }
    MutexT *stealLock()
    { MutexT *ret = mutex; mutex = 0; return ret; }
  private:
    BasicMutexLocker(const BasicMutexLocker &);
    BasicMutexLocker &operator= (const BasicMutexLocker &);
  };

  typedef BasicMutexLocker<Mutex> MutexLocker;
  typedef BasicMutexLocker<FileMutex> FileMutexLocker;

  class FANCYBASE_API WaitCondition
  {
  public:
    WaitCondition();
    ~WaitCondition();

    void broadcast();
    void signal();

    void wait (Mutex *mutex);
    bool timedWait(Mutex *mutex, unsigned msec);
  private:
    pthread_cond_t cond;
  };

  class FANCYBASE_API RWMutex
  {
  public:
    RWMutex();
    ~RWMutex();

    void rLock();
    void rUnlock();

    bool wLockTimed(int msec);
    void wLock();
    void wUnlock();

  private:
    int n_read;
    Mutex n_read_mutex;
    WaitCondition n_read_wc;
    Mutex write_mutex;
  };

  class FANCYBASE_API RMutexLocker
  {
    RWMutex *mutex;
  public:
    RMutexLocker(RWMutex *m) : mutex(m) { if (mutex) mutex->rLock(); }
    ~RMutexLocker() { if (mutex) mutex->rUnlock(); }
    void unlock() { if (mutex) { mutex->rUnlock(); mutex=0; } }
    void lock(RWMutex *m)
    { if (mutex) mutex->rUnlock(); mutex=m; if (mutex) mutex->rLock(); }
  private:
    RMutexLocker(const RMutexLocker &);
    RMutexLocker &operator= (const RMutexLocker &);
  };

  class FANCYBASE_API WMutexLocker
  {
    RWMutex *mutex;
  public:
    WMutexLocker() : mutex(0) {}
    WMutexLocker(RWMutex *m) : mutex(m) { if (mutex) mutex->wLock(); }
    ~WMutexLocker() { if (mutex) mutex->wUnlock(); }
    void unlock() { if (mutex) { mutex->wUnlock(); mutex=0; } }
    void lock(RWMutex *m)
    { if (mutex) mutex->wUnlock(); mutex=m; if (mutex) mutex->wLock(); }
    bool isLocked() const { return mutex!=0; }
    bool lockTimed(RWMutex *m, int msec)
    { if (mutex) mutex->wUnlock(); mutex=0; if (m && m->wLockTimed(msec)) mutex = m; return isLocked(); }
  private:
    WMutexLocker(const WMutexLocker &);
    WMutexLocker &operator= (const WMutexLocker &);
  };

  /*! @class Thread

    @brief Encapsulates a pthread thread.

    This class encapsulates a joinable POSIX thread.
    The main loop is represented by the Runnable instance passed
    to @c newInstance().

    Subclassing the thread class is deprecated and will not work
    as of ABI version 2.
  */
  class FANCYBASE_API Thread : virtual public Object
  {
    friend class ThreadStarter;

  public:
#ifdef WIN32
    typedef unsigned retval_t;
#else
    typedef void *retval_t;
#endif
    typedef ObjPtr<Thread> ptr_t;

    static ptr_t newInstance(const Runnable::ptr_t &runnable);

    virtual void kill(int sig) = 0;

    /**
      Waits for the termination of the thread launched by start().

      Returns the exit status of thread:

      @li FANCYBASE_THREAD_SUCCESS  The thread terminated without an uncaught exception.
      @li FANCYBASE_THREAD_EXCEPTION
      The thread terminated by catching a Fancy::Exception.
      @li FANCYBASE_THREAD_STD_EXCEPTION
      The thread terminated by catching a std::exception.
      @li FANCYBASE_THREAD_UNKNOW_EXCEPTION
      The thread terminated by catching an unknown exception.
      @li Otherwise, the macro FANCYBASE_THREAD_ERRNO(a) may be used to retrieve
      the libc errno of any unhandled error inside a C std library call.
    */
    virtual retval_t wait() = 0;

    /**
       Launches the new thread. The new thread calls the
       @c Runnable passed to @c newInstance().
     */
    virtual void start() = 0;

    /**
     * Return whether this thread object represents the current thread.
     */
    virtual bool isSelf() const = 0;

    /**
     * Returns, whether this thread has been started and @c wait()
     * has not yet been called.
     */
    virtual bool isRunning() const = 0;

    virtual Runnable::ptr_t getRunnable() const = 0;

    virtual pthread_t handle() const = 0;
    static pthread_t currentHandle();

    virtual long getId() const = 0;
    static long getCurrentId();

#ifdef WIN32
    static void *getSignalWinEventHandle();
    static void *getIOWinEventHandle();
    static int getSignalCatched();
#endif
    static void msleep (unsigned msec);

    static Object::ptr_t getSpecific(const Object::ptr_t &key);
    static void setSpecific(const Object::ptr_t &key, const Object::ptr_t &value);
    static Object::ptr_t removeSpecific(const Object::ptr_t &key);

    /**
     * Register a thread callback. Please note, that this function is not thread-safe.
     * It is assumed, that thread callbacks are registered in the main program before
     * any threads are actually started.
     */
    static void registerCallback(const ObjPtr<ThreadCallback> &cb);

    /**
     * Deregister a thread callback. Please note, that this function is not thread-safe.
     * It is assumed, that thread callbacks are deregistered in the main program after
     * any threads have actually been stopped.
     */
    static void deregisterCallback(const ObjPtr<ThreadCallback> &cb);

  protected:

    Thread();
    virtual ~Thread();

  private:
  };

#define FANCYBASE_THREAD_SUCCESS          ((::Fancy::Thread::retval_t)0)
#define FANCYBASE_THREAD_EXCEPTION        ((::Fancy::Thread::retval_t)-1)
#define FANCYBASE_THREAD_STD_EXCEPTION    ((::Fancy::Thread::retval_t)-2)
#define FANCYBASE_THREAD_UNKNOW_EXCEPTION ((::Fancy::Thread::retval_t)-3)
#define FANCYBASE_THREAD_ERRNO(a)         ((int)(a))

  /**
   * A callback to executed before and after Thread::run() regardless
   * whether the thread threw an exception or not.
   *
   * Please note, that a ThreadCallback is not executed, if a thread is not
   * started using Fancy::Thread but throw an operating system primitive.
   */
  class FANCYBASE_API ThreadCallback : virtual public Object {

  public:
      typedef ObjPtr<ThreadCallback> ptr_t;

      /**
       * Called right before Thread::run().
       */
      virtual void start() = 0;

      /**
       * Called right before Thread::run().
       *
       * @param ret The return value of the thread.
       */
      virtual void leave(Thread::retval_t ret) = 0;

  protected:
      ThreadCallback() {}
      ~ThreadCallback();
  };

  /**
   * A ctor/dtor based scope for setting a thread-specific variable using
   * \c Thread::setSpecific(key,value) for a specified scope.
   */
  class ThreadLocalContext {
	Object::ptr_t key;
	
  public:
    ThreadLocalContext(const Object::ptr_t &_key, const Object::ptr_t &value) {

	  if (!value.isNull()) {
		this->key = _key;
	  }

	  if (!this->key.isNull()) {
		Thread::setSpecific(this->key,value);
	  }
	}

    ~ThreadLocalContext() {
	  
	  if (!this->key.isNull()) {
		Thread::removeSpecific(this->key);
	  }
	}

  };
}

#endif // ! FANCYBASE_THREAD_H
