/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2009-2011 ev-i GmbH & IBCL e.U.
 * Copyright (C) 2011-2013 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCYUID_H_
#define FANCYUID_H_

#include <fancyserializable.h>

namespace Fancy {

/**
 * A unique identifier, usually constructed from hostname, process Id and the time of creation.
 */
class FANCYBASE_API UID: public virtual Fancy::Serializable {

    UID(uint32_t u, uint64_t t, uint16_t c);
    UID();

public:
    typedef ObjPtr<UID> ptr_t;

    /**
     * Return an uninitialized instance, which will be initialized through
     * @c Serializable::readFields().
     */
    static ObjPtr<Serializable> newSerializableInstance();

    /**
     * Return an instance, which is initialized in a way, that it's fields are unique.
     * Uniqueness is defined by the hostname, process Id and the time of creation of the
     * UID measured in millseconds since the epoch 1970-01-01 00:00:00 GMT.
     */
    static ptr_t newInstance();

	// override.
	std::string toString() const;

	bool operator== (const UID &a) const;
	bool operator!= (const UID &a) const;
    bool operator<  (const UID &a) const;

    bool lessThan(const Object &o) const;
    bool equals(const Object &o) const;

    /*!
     * Return the unique part of the UID, which is derived from the hostname and the
     * process ID.
     */
    uint32_t getUnique() const { return unique; }

    /*!
     * Return the reference time of the UID, which is the number milliseconds
     * since the epoch 1970-01-01 00:00:00 GMT of the first generated UID in
     * a row.
     */
    uint64_t getTime()   const { return time; }

    /*!
     * Return the counter, which is incremented for each UID. If the counter overflows,
     * a new timestamp is generated.
     */
    uint16_t getCount()  const { return this->count; }

    // implementation of Fancy::Serializable
    static int64_t getSerialVersionUID_s();
    static const char *getSerialID_s();

    int64_t getSerialVersionUID() const;
    const char *getSerialID() const;
    void writeFields(const ObjPtr<Serializer> &serializer) const;
    ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);


protected:
	~UID();

private:
	// this type is immutable.
	UID &operator=(const UID &);

	uint32_t unique;
	uint64_t time;
	uint16_t count;
};

}

#endif /* FANCYUID_H_ */
